import sys
from setuptools import setup, find_packages

## Tag
#
# To create a tag in the repo
#
# ```bash
#     git tag X.Y.Z -m "pycollector-X.Y.Z"
#     git push --tags origin master
# ```
#
## PyPI distribution
#
# ```bash
# python3 setup.py sdist upload -r pypi
# ```

d_version = {}
with open("./pycollector/version.py") as fp:
    exec(fp.read(), d_version)
version = d_version['VERSION']

d_setup = {'author':'Visym Labs',
           'author_email':'info@visym.com',
           'version':version,
           'namespace_packages':['pycollector'],
           'packages':find_packages(),
           'description':'Visym Collector',
           'long_description':"Visym Collector Python Tools for Live Visual Datasets",
           'long_description_content_type':"text/markdown",
           'url':'https://github.com/visym/collector',
           'download_url':'https://github.com/visym/collector/archive/%s.tar.gz' % version,
           'install_requires':["vipy","dask","distributed","boto3","xmltodict","pandas","torch","torchvision","pytorch_lightning","ujson"],
           #'dependency_links':[]  # FIXME: there is an issue with windows installs of torch
           'keywords':['computer vision machine learning ML CV privacy video image'],
           'classifiers':["Programming Language :: Python :: 3",
                          "Operating System :: OS Independent",
                          "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"],
           'include_package_data':True}



import os
import boto3
import botocore
from botocore.config import Config
import requests
from datetime import datetime, timedelta
import decimal
import pandas as pd
import numpy as np
import json

# import logging
import pycollector
import getpass
from pycollector.globals import GLOBALS  # print
from pycollector.util import is_email_address


class User(object):
    """User class for collector's user management"""

    def __init__(self, username=None, password=None):
        """

        Args:
            username ([str]): username of pycollector.
            password ([str], optional): password for the user. Defaults to None.
        """
        # TODO - Apply paramter store

        # self.app_client_id = GLOBALS["COGNITO"]["app_client_id"]
        # self.identity_pool_id = GLOBALS["COGNITO"]["identity_pool_id"]
        # self.provider_name = GLOBALS["COGNITO"]["provider_name"]
        # self.region_name = GLOBALS["COGNITO"]["region_name"]

        # Initialize the base user properties and create the logger used by the function
        self._is_login = False
        # self._token_initialized_time = None
        self._token_expiration_time = None
        self._program_name = None

        # # Initialize cognito clients
        # # Ensure the system AWS credentials are not being used
        # config = Config(signature_version=botocore.UNSIGNED)
        # self._cognito_idp_client = boto3.client("cognito-idp", config=config, region_name=self.region_name)
        # self._cognito_id_client = boto3.client("cognito-identity", config=config, region_name=self.region_name)

        # Login
        if username is None and "VISYM_COLLECTOR_EMAIL" in os.environ:
            username = os.environ["VISYM_COLLECTOR_EMAIL"]
        # Set user properties
        self._username = username
        self._password = password
        if password is not None:
            self.login(password)

        self.region_name = 'us-east-1'
        self.refresh()

    def refresh(self):
        if "VIPY_AWS_SESSION_TOKEN" in os.environ:
            self._set_S3_clients()
            self._set_lambda_clients()
            self._is_login = True
            self._cognito_username = os.environ["VIPY_AWS_COGNITO_USERNAME"]
        return self

    def login(self, password=None):
        """[summary]

        Args:
            username ([str]): username of pycollector.
            password ([str], optional): password for the user. Defaults to None.
        """

        username = self._username if self._username is not None else input("Collector email: ")
        assert is_email_address(username), 'Invalid collector email address "%s"' % username
        password = password if password is not None else getpass.getpass()

        try:
            # Set up API gateway request for login
            request_body = {"username": username, "password": password}
            self._aws_credentials = requests.post(
                GLOBALS["API_GATEWAY_HTTP"]["pycollector_login"],
                data=json.dumps(request_body),
                headers={"Content-type": "application/json", "Accept": "text/plain"},
            ).json()

            self._cognito_username = self._aws_credentials["cognito_username"]
            self._token_expiration_time = datetime.now() + timedelta(0, self._aws_credentials["token_expires_in_secs"])
            self.region_name = self._aws_credentials["region_name"]

            # Set up AWS services
            self._set_os_environ()
            self._set_parameter_store()
            self._set_S3_clients()
            self._set_lambda_clients()
            self._is_login = True

        except Exception as e:
            custom_error = "Failed to sign in due to exception: {0}".format(e)
            raise Exception(custom_error)

        return self

    def _set_parameter_store(self):
        """[summary]"""
        self._ssm_client = boto3.client(
            "ssm",
            aws_access_key_id=os.environ["VIPY_AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["VIPY_AWS_SECRET_ACCESS_KEY"],
            aws_session_token=os.environ["VIPY_AWS_SESSION_TOKEN"],
            region_name=os.environ["VIPY_AWS_REGION"],
        )

    def get_ssm_param(self, param_name: str = None, WithDecryption: bool = False) -> str:
        """[summary]"""
        self._set_parameter_store()
        return self._ssm_client.get_parameter(Name=param_name, WithDecryption=WithDecryption).get("Parameter").get("Value")

    def _set_S3_clients(self):
        """[summary]"""
        assert "VIPY_AWS_SESSION_TOKEN" in os.environ
        self._s3_client = boto3.client(
            "s3",
            aws_access_key_id=os.environ["VIPY_AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["VIPY_AWS_SECRET_ACCESS_KEY"],
            aws_session_token=os.environ["VIPY_AWS_SESSION_TOKEN"],
            region_name=os.environ["VIPY_AWS_REGION"],
        )
        self._s3_resource = boto3.resource(
            "s3",
            aws_access_key_id=os.environ["VIPY_AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["VIPY_AWS_SECRET_ACCESS_KEY"],
            aws_session_token=os.environ["VIPY_AWS_SESSION_TOKEN"],
            region_name=os.environ["VIPY_AWS_REGION"],
        )

    def _set_lambda_clients(self):
        """[summary]"""
        assert "VIPY_AWS_SESSION_TOKEN" in os.environ
        self._lambda_client = boto3.client(
            "lambda",
            aws_access_key_id=os.environ["VIPY_AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["VIPY_AWS_SECRET_ACCESS_KEY"],
            aws_session_token=os.environ["VIPY_AWS_SESSION_TOKEN"],
            region_name=os.environ["VIPY_AWS_REGION"],
        )

    def _set_os_environ(self):
        """[summary]"""
        os.environ["VIPY_AWS_ACCESS_KEY_ID"] = self._aws_credentials["access_key_id"]
        os.environ["VIPY_AWS_SECRET_ACCESS_KEY"] = self._aws_credentials["secret_key"]
        os.environ["VIPY_AWS_SESSION_TOKEN"] = self._aws_credentials["session_token"]
        os.environ["VIPY_AWS_COGNITO_USERNAME"] = self._cognito_username
        os.environ["VIPY_AWS_REGION"] = self._aws_credentials["region_name"]

    def is_token_expired(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        return self._token_expiration_time is None or datetime.now() > self._token_expiration_time

    def token_expired_by(self):
        return self._token_expiration_time

    def is_authenticated(self):
        return self._is_login

    def add_user_to_group(self):
        """Check if the current user is already in the pycollector user group, if not add the user to group

        Returns:
            [type]: [description]
        """

    @property
    def username(self):
        """"""
        return self._username

    @property
    def cognito_username(self):
        """"""
        return self._cognito_username

    @property
    def lambda_client(self):
        """"""
        return self._lambda_client



from models.bases.base import Base
from models.bases.resnet50_3d import ResNet3D, Bottleneck3D


class ResNet3DEncoder(ResNet3D):
    def __init__(self, block, layers, num_classes=400):
        super(ResNet3DEncoder, self).__init__(block, layers, num_classes)
        self.fc = None

    def forward(self, x):
        # x is of the form b x n x h x w x c
        # model expects b x c x n x h x w
        x = x.permute(0, 4, 1, 2, 3)

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.maxpool2(x)
        x = self.layer2(x)
        #x = self.layer3(x)
        #x = self.layer4(x)

        #x = self.avgpool(x)
        #x = self.dropout(x)
        #logits = self.fc(x)

        #logits = logits.mean(3).mean(3)
        # model returns batch x classes x time
        #logits = logits.permute(0, 2, 1)
        # logits is batch X time X classes
        #if self.global_pooling:
        #    logits = logits.mean(1)
        #return logits
        return x

class ResNet503DEncoder(Base):
    @classmethod
    def get(cls, args):
        model = ResNet3DEncoder(Bottleneck3D, [3, 4, 6, 3])  # 50
        if args.pretrained:
            from torchvision.models.resnet import resnet50
            model2d = resnet50(pretrained=True)
            model.load_2d(model2d)
        return model



import json
import pandas as pd
import ast
from pycollector.user import User
from pycollector.project import Project
from pycollector.globals import print, GLOBALS


class API(User):
    """class for pycollector API

    Args:
        object ([type]): [description]
    """

    def __init__(self, username=None, password=None):
        """

        Args:
            username ([type], optional): [description]. Defaults to None.
            password ([type], optional): [description]. Defaults to None.
        """
        super().__init__(username=username, password=password)
        if not self.is_authenticated():
            self.login()

    def get_project(self,
                    project=None,
                    alltime=True,
                    weeksago=None,
                    monthsago=None,
                    daysago=None,
                    since=None,
                    before=None,
                    last=None,
                    retry=2,
    ):
        """[summary]

        Args:
            -
        """

        return Project(
            project=project,
            alltime=alltime,
            weeksago=weeksago,
            monthsago=monthsago,
            daysago=daysago,
            since=since,
            before=before,
            last=last,
            retry=retry,
            username=self._username,
            password=self._password,
        )

    def videos(self, n):
        """[summary]

        Args:
            -
        """

        return Project(last=n).last(n)

    def lastvideo(self):
        return self.videos(n=1)
    
    def new_collection(
            self,
            name,
            # program_name,
            project_name,
            collection_description,
            activities,
            activity_short_names,
            actor,
            consent_overlay_text='Please select the record button, say "I consent to this video collection”',
            objects=None,            
    ):

        assert isinstance(name, str)
        assert isinstance(project_name, str)
        assert isinstance(collection_description, str)
        assert isinstance(activities, str) or (isinstance(activities, tuple) and all([isinstance(a, str) for a in activities]))
        assert isinstance(activity_short_names, str) or (isinstance(activity_short_names, tuple) and all([isinstance(a, str) for a in activity_short_names]))
        assert (isinstance(activities, str) and isinstance(activity_short_names, str)) or len(activities) == len(activity_short_names)
        assert isinstance(actor, str)
        assert objects is None or isinstance(objects, str) or (isinstance(objects, tuple) and all([isinstance(o, str) for o in objects]))
                                               
        activities_csv = ','.join(activities) if isinstance(activities, tuple) else activities
        activity_short_names_csv = ','.join(activity_short_names) if isinstance(activity_short_names, tuple) else activity_short_names
        objects = () if objects is None else objects
        objects_csv = ','.join(objects) if isinstance(objects, tuple) else objects
        
        request = {
            "collection_name": name,
            "activities": activities_csv,  # comma separated string, no spaces
            "project_name": project_name,
            # "program_name": program_name,
            "objects": objects_csv,  # comma separated string, no spaces
            "actor": actor,
            "collection_description": collection_description,
            "activity_short_names": activity_short_names_csv,  # comma separated string, no spaces
            "creator_cognito_username": self.cognito_username,
            "consent_overlay_text": consent_overlay_text,  # currently disabled
        }

        # Invoke Lambda function
        FunctionName = self.get_ssm_param(GLOBALS["LAMBDA"]["create_collection"])
        response = self._lambda_client.invoke(
            FunctionName=FunctionName,
            InvocationType="RequestResponse",
            LogType="Tail",
            Payload=json.dumps(request),
        )

        d = json.loads(response['Payload'].read().decode('UTF-8'))
        if 'statusCode' not in d or d['statusCode'] != 200:
            raise ValueError(str(d))
        else:
            return d

    def list_collections(self):

        # Invoke Lambda function
        request = {"collectorid": self.cognito_username}

        # Invoke Lambda function
        FunctionName = self.get_ssm_param(GLOBALS["LAMBDA"]["list_collections"])
        response = self._lambda_client.invoke(
            FunctionName=FunctionName,
            InvocationType="RequestResponse",
            LogType="Tail",
            Payload=json.dumps(request),
        )
        
        # Get the serialized dataframe
        dict_str = response["Payload"].read().decode("UTF-8")
        if dict_str == "null":
            raise ValueError("Invalid lambda function response")
        data_dict = ast.literal_eval(dict_str)
        
        serialized_collections_data_dict = data_dict["body"]["collections"]

        # Return list of dictionaries
        h = ['created_date', 'name', 'collection_id', 'project_name']
        d = {k:v for (k,v) in json.loads(serialized_collections_data_dict).items() if k in h}
        n = len(d['name']) if 'name' in d else 0
        return [{k:v[str(j)] for (k,v) in d.items()} for j in range(0,n)]


    def delete_collection(self, collectionid):

        assert isinstance(collectionid, str)
        
        # Invoke Lambda function
        request = {"collectorid": self.cognito_username, "collectionid": collectionid}

        # Invoke Lambda function
        FunctionName = self.get_ssm_param(GLOBALS["LAMBDA"]["delete_collection"])
        response = self._lambda_client.invoke(
            FunctionName=FunctionName,
            InvocationType="RequestResponse",
            LogType="Tail",
            Payload=json.dumps(request),
        )
        
        # Get the serialized message
        d = json.loads(response['Payload'].read().decode('UTF-8'))
        if 'statusCode' not in d or d['statusCode'] != 200:
            raise ValueError(str(d))
        else:
            return d
            
